module Lava where



{- Erster Versuch: Flache Einbettung: Signale sind (unendliche) Folgen
   von Bits (True = 1, False = 0). -}
newtype S_Signal = S_Sig [Bool]
un_S_Sig (S_Sig x) = x
instance Show S_Signal where
  show a = '$' : map (\ x-> if x then '1' else '0') (un_S_Sig a)
  
s_lift :: (Bool -> Bool) -> S_Signal -> S_Signal
s_lift f = S_Sig . map f . un_S_Sig

s_lift2 :: ((Bool, Bool) -> Bool) -> (S_Signal, S_Signal) -> S_Signal
s_lift2 f (as, bs) = S_Sig $ zipWith (curry f) (un_S_Sig as) (un_S_Sig bs)

{- Die ueblichen binaeren Schaltkreise -}
s_and, s_or, s_xor, s_nand :: (S_Signal, S_Signal) -> S_Signal

s_and ab = (s_lift2 $ uncurry (&&)) ab
s_or ab = (s_lift2 $ uncurry (||)) ab
s_xor ab = (s_lift2 (\ (x, y)-> if x then not y else y)) ab
s_nand ab = (s_lift2 (\ (x, y)-> not (x && y))) ab


{- Inversion eines Signals -}
s_inv :: S_Signal -> S_Signal
s_inv as = s_lift not as



{- Ein erster Schaltkreis: Halbaddierer. Wir sehen:
   mehrere Ausgaenge werden (hier) durch Tupel repraesentiert. -}
s_halfAdder :: (S_Signal, S_Signal) -> (S_Signal, S_Signal)
s_halfAdder (a, b) = (sum, carry)
  where sum = s_xor (a, b)
        carry = s_and (a, b)

{- Eine Folge von '0' bzw. '1'. -}
s_low, s_high :: S_Signal
s_low = S_Sig $ repeat False
s_high = S_Sig $ repeat True

s_mksig :: [Bool] -> S_Signal
s_mksig bs = S_Sig bs

{- In einer flachen Einbettung bekommen wir den Interpreter gemeinhin
   geschenkt, da alle Kombinatoren bereits auf Domaenenwerten arbeiten.
   Simulation eines Schaltkreises bedeutet also einfach die Entnahme
   der anfaenglichen N Bits. 
   
   Wir wollen jedoch nicht nur S_Signal, also [Bool] simulieren, 
   sondern auch Tupel und Listen hiervon
   -}
   
class S_Sim a where
  s_sim :: Int -> a -> [a]

instance S_Sim S_Signal where
  s_sim n = take n . map (S_Sig . (:[])) . un_S_Sig

-- ([a1,...], [b1,...]) nach [(a1, b1), (a2, b2), ...]
instance (S_Sim a, S_Sim b)=> S_Sim (a, b) where
  s_sim n (a, b) = zipWith (,) (s_sim n a) (s_sim n b)
  
-- [[a1,...], [b1,...], [c1,...]] nach [[a1, b1, c1], [a2, b2, c2], ...]
instance S_Sim a=> S_Sim [a] where
  s_sim n as = let a123s = map (s_sim n) as
                   f ([]:_) = []
                   f xs = map head xs : f (map tail xs)
               in f a123s


s_fullAdder (carryIn, (a, b)) = (sum, carryOut)
  where (sum1, carry1) = s_halfAdder (a, b)
        (sum, carry2) = s_halfAdder (carryIn, sum1)
        carryOut = s_xor (carry1, carry2)


{- Ein Bitsortierer -}        
s_twoBitSorter (a, b) = (a', b')
  where a' = s_and (a, b)
        b' = s_or (a, b)

        
s_bitAdder :: (S_Signal, [S_Signal]) -> ([S_Signal], S_Signal)
s_bitAdder (carryIn, []) = ([], carryIn)
s_bitAdder (carryIn, (a:as)) = (sum:sums, carryOut)
  where (sum, carry) = s_halfAdder (carryIn, a)
        (sums, carryOut) = s_bitAdder (carry, as)



-- Wrong: s_adder :: (S_Signal, [S_Signal]) -> ([S_Signal], S_Signal)
s_adder (carryIn, ([], [])) = ([], carryIn)
s_adder (carryIn, (a:as, b:bs)) = (sum:sums, carryOut)
  where (sum, carry) = s_fullAdder (carryIn, (a, b))
        (sums, carryOut) = s_adder (carry, (as, bs))

        
        

s_row circ (carryIn, []) = ([], carryIn)
s_row circ (carryIn, a:as) = (b:bs, carryOut)
  where (b, carry) = circ (carryIn, a)
        (bs, carryOut) = s_row circ (carry, as)

{- Ein Addierer, jetzt generisch definierbar (wenn auch mit leicht
   veraendertem Typen). -}        
s_adder' as = s_row s_fullAdder as





{- Essentiell fuer rueckgekoppelte Schaltkreise: 
   Delays (a.k.a. Flip-flops). Der initiale Ausgang des Delays
   wird als Parameter angegeben. -}
s_delay :: Bool -> S_Signal -> S_Signal
s_delay init as = S_Sig $ init : un_S_Sig as



{- Ein Flankendetektor -}
s_edge a = change
  where a' = s_delay False a
        change = s_xor (a, a')


{- Rekursion, bzw. eine Schleife im Schaltkreis: 
   ein Zustands-Delay, das seinen Wert aendert, 
   wenn die Eingabe '1' ist -}
s_toggle change = out
  where out' = s_delay False out
        out = s_xor (change, out')



s_cnt () = number'
  where number' = s_delay False number
        (number, _) = s_halfAdder(s_high, number')

-- [Sig [False, ...], Sig [False, ...], ... n times, Sig [False, ...]]
s_delayNF n sigs = 
  if n == 0 then []
    else S_Sig (False : un_S_Sig (head sigs)) : s_delayNF (n - 1) (tail sigs)

s_counter n () = number'
  where number' = s_delayNF n number
        (number, carryOut) = s_bitAdder (s_high, number')

{-
-- Geht so nicht: zipWith muss @as@ ablaufen koennen, um Fortschritt zu 
-- machen. @as@ ist aber nur dann bestimmt, wenn delayN ohne Auswertung
-- von @number@, also @as@ Fortschritt macht. Zyklus! 
s_counter n () = number'
  where delayN inits as = zipWith ($) (map s_delay inits) as
        zeroList n = replicate n False
        number' = delayN (zeroList n) $ number
        (number, carryOut) = s_bitAdder (s_high, number')
-}



{-
  Lava, Teil II:
  Die flache Einbettung erlaubt es nicht, den erzeugten Schaltkreis
  zu durchlaufen, um etwa eine Netlist, d.h. eine grafische Darstellung
  zu erzeugen. Hierfuer ist eine tiefe Einbettung noetig, d.h. wir
  brauchen eine Darstellung von Schaltkreisen als Terme, und koennen
  nicht einfach auf der `Semantik' als [Bool] arbeiten.
-}

{- Hier starke Vereinfachung: Ein Signal ist entweder ein Eingang
   (Var, mit einem eindeutigen Namen versehen) oder eine Komponente,
   die verschiedene Eingaenge (Signal) erwartet. Ihre Ausgaenge
   werden durch den jetzt zu implementierenden Interpreter bestimmt.
-}
data Signal = Var String | Comp String [Signal]
  deriving Show
  
d_inv b = Comp "inv" [b]
d_delay b = Comp "delay" [b]
d_and (a, b) = Comp "and" [a, b]
d_xor (a, b) = Comp "xor" [a, b]

d_halfAdder (a, b) = (sum, carry)
  where sum = d_xor (a, b)
        carry = d_and (a, b)
                

d_toggle = output
  where output = d_inv (d_delay output)

d_interp _ [] = []
d_interp (Var x) (vs:z) = (case lookup x vs of
                             Nothing -> False
                             Just v -> v) : d_interp (Var x) z
d_interp b@(Comp op sigs) g@(vs:z) = 
  (case op of
    "inv" -> let [v] = map (\ s-> d_interp s g) sigs 
             in map not v
    "and" -> let [v, v'] = map (\ s-> d_interp s g) sigs
             in zipWith (&&) v v'
    "xor" -> let [v, v'] = map (\ s-> d_interp s g) sigs
             in zipWith (\ v v'-> (if v then not v' else v')) v v'
    "delay" -> False : (let [v] = map (\ s-> d_interp s g) sigs in v))
    
    
type Id = Int
data Signal' = Var' String | Comp' String Id [Signal]
